# exports the name of the training algorithm
from pathlib import Path

import numpy as np
from gymnasium import spaces
from negmas.sao.common import SAOResponse
from scml.std.awi import StdAWI
from scml.std.common import StdState
from scml.std.rl.context import LimitedPartnerNumbersStdContext, WorldContext
from scml.std.rl.observation import (
    LimitedPartnerNumbersObservationManager,
    ObservationManager,
)
from stable_baselines3 import A2C
from stable_baselines3.common.base_class import BaseAlgorithm

N_AGENTS = (1, 8)
"""Numbers of agents per level supported"""

TrainingAlgorithm: type[BaseAlgorithm] = A2C
"""The algorithm used for training. You can use any stable_baselines3 algorithm or develop your own"""

MODEL_PATH = Path(__file__).parent / "models" / "mymodel"
"""The path in which train.py saves the trained model and from which myagent.py loads it"""


ObsType: type[ObservationManager] = LimitedPartnerNumbersObservationManager
"""Observation manager base class"""

ContextType: type[WorldContext] = LimitedPartnerNumbersStdContext
"""A world context that can generate any world compatible with the observation manager"""


def make_context(level: int) -> ContextType:
    """Generates a context as a supplier or as a consumer"""

    if level == 0:
        return ContextType(
            n_suppliers=(0, 0),  # suppliers have no suppliers
            n_consumers=N_AGENTS,
            n_competitors=(N_AGENTS[0] - 1, N_AGENTS[1] - 1),
            n_agents_per_level=N_AGENTS,
            level=0,  # suppliers are always in the first level
        )
    elif level == -1:
        return ContextType(
            n_suppliers=N_AGENTS,
            n_consumers=(0, 0),  # consumers have no consumers
            n_competitors=(N_AGENTS[0] - 1, N_AGENTS[1] - 1),
            n_agents_per_level=N_AGENTS,
            level=-1,  # consumers are always in the last level
        )
    return ContextType(
        n_suppliers=N_AGENTS,
        n_consumers=N_AGENTS,
        n_competitors=(N_AGENTS[0] - 1, N_AGENTS[1] - 1),
        n_agents_per_level=N_AGENTS,
        level=level,  # consumers are always in the last level
    )


class MyObservationManager(ObsType):
    """This is my observation manager implementing encoding and decoding the state used by the RL algorithm"""

    def make_space(self) -> spaces.Space:
        """Creates the observation space"""
        return super().make_space()

    def encode(self, state: StdState) -> np.ndarray:
        """Encodes an observation from the agent's state"""
        return super().encode(state)

    def make_first_observation(self, awi: StdAWI) -> np.ndarray:
        """Creates the initial observation (returned from gym's reset())"""
        return super().make_first_observation(awi)

    def get_offers(self, awi: StdAWI, encoded: np.ndarray) -> dict[str, SAOResponse]:
        """Gets the offers from an encoded state"""
        return super().get_offers(awi, encoded)


# ensure that the folder containing models is created
MODEL_PATH.parent.mkdir(parents=True, exist_ok=True)
